from jax.lax.linalg import (
    cholesky as cholesky,
    eig as eig,
    eigh as eigh,
    hessenberg as hessenberg,
    householder_product as householder_product,
    lu as lu,
    qdwh as qdwh,
    qr as qr,
    shur as shur,
    svd as svd,
    triangular_solve as triangular_solve,
    tridiagonal as tridiagonal,
    tridiagonal_solve as tridiagonal_solve,
)



"""Quaxified `jax.numpy`.

This module wraps the functions in `jax.numpy` with `quax.quaxify`. The wrapping
happens dynamically through a module-level ``__dir__`` and ``__getattr__``. The
list of available functions is in ``__all__`` and documented in the `jax.numpy`
library.

In addition the following modules are supported:

- `quaxed.numpy.fft`
- `quaxed.numpy.linalg`

The contents of these modules are likewise dynamically wrapped with
`quax.quaxify` and their contents is listed in their respective ``__all__`` and
documented in their respective libraries.

If a function is missing, please file an Issue.

"""
# pylint: disable=redefined-builtin

from typing import Any

from jaxtyping import install_import_hook

from . import _core, _creation_functions, _higher_order, fft, linalg
from ._creation_functions import *
from ._higher_order import *

__all__ = (
    ["fft", "linalg"]  # noqa: RUF005
    + _core.__all__
    + _higher_order.__all__
    + _creation_functions.__all__
)


# TODO: consolidate with ``_core.__getattr__``.
def __getattr__(name: str) -> Any:
    if name in __all__:
        return getattr(_core, name)

    msg = f"module {__name__!r} has no attribute {name!r}"
    raise AttributeError(msg)


# TODO: figure out how to install this import hook, with the __getattr__.
install_import_hook("quaxed.numpy", None)



from jax.scipy.special import (
    bernoulli as bernoulli,
    bessel_jn as bessel_jn,
    beta as beta,
    betainc as betainc,
    betaln as betaln,
    digamma as digamma,
    entr as entr,
    erf as erf,
    erfc as erfc,
    erfinv as erfinv,
    exp1 as exp1,
    expi as expi,
    expit as expit,
    expn as expn,
    factorial as factorial,
    gamma as gamma,
    gammainc as gammainc,
    gammaincc as gammaincc,
    gammaln as gammaln,
    hyp1f1 as hyp1f1,
    i0 as i0,
    i0e as i0e,
    i1 as i1,
    i1e as i1e,
    kl_div as kl_div,
    log_ndtr as log_ndtr,
    logit as logit,
    logsumexp as logsumexp,
    lpmn as lpmn,
    lpmn_values as lpmn_values,
    multigammaln as multigammaln,
    ndtr as ndtr,
    ndtri as ndtri,
    poch as poch,
    polygamma as polygamma,
    rel_entr as rel_entr,
    spence as spence,
    sph_harm as sph_harm,
    xlog1py as xlog1py,
    xlogy as xlogy,
    zeta as zeta,
)



# ----- Operators -----
# isort: split
from jax.lax import (
    abs as abs,
    acos as acos,
    acosh as acosh,
    add as add,
    approx_max_k as approx_max_k,
    approx_min_k as approx_min_k,
    argmax as argmax,
    argmin as argmin,
    asin as asin,
    asinh as asinh,
    atan as atan,
    atan2 as atan2,
    atanh as atanh,
    batch_matmul as batch_matmul,
    bessel_i0e as bessel_i0e,
    bessel_i1e as bessel_i1e,
    betainc as betainc,
    bitcast_convert_type as bitcast_convert_type,
    bitwise_and as bitwise_and,
    bitwise_not as bitwise_not,
    bitwise_or as bitwise_or,
    bitwise_xor as bitwise_xor,
    broadcast as broadcast,
    broadcast_in_dim as broadcast_in_dim,
    broadcast_shapes as broadcast_shapes,
    broadcast_to_rank as broadcast_to_rank,
    broadcasted_iota as broadcasted_iota,
    cbrt as cbrt,
    ceil as ceil,
    clamp as clamp,
    clz as clz,
    collapse as collapse,
    complex as complex,
    concatenate as concatenate,
    conj as conj,
    conv as conv,
    conv_dimension_numbers as conv_dimension_numbers,
    conv_general_dilated as conv_general_dilated,
    conv_general_dilated_local as conv_general_dilated_local,
    conv_general_dilated_patches as conv_general_dilated_patches,
    conv_transpose as conv_transpose,
    conv_with_general_padding as conv_with_general_padding,
    convert_element_type as convert_element_type,
    cos as cos,
    cosh as cosh,
    cumlogsumexp as cumlogsumexp,
    cummax as cummax,
    cummin as cummin,
    cumprod as cumprod,
    cumsum as cumsum,
    digamma as digamma,
    div as div,
    dot as dot,
    dot_general as dot_general,
    dynamic_index_in_dim as dynamic_index_in_dim,
    dynamic_slice as dynamic_slice,
    dynamic_slice_in_dim as dynamic_slice_in_dim,
    dynamic_update_index_in_dim as dynamic_update_index_in_dim,
    dynamic_update_slice as dynamic_update_slice,
    dynamic_update_slice_in_dim as dynamic_update_slice_in_dim,
    eq as eq,
    erf as erf,
    erf_inv as erf_inv,
    erfc as erfc,
    exp as exp,
    expand_dims as expand_dims,
    expm1 as expm1,
    fft as fft,
    floor as floor,
    full as full,
    full_like as full_like,
    gather as gather,
    ge as ge,
    gt as gt,
    igamma as igamma,
    igammac as igammac,
    imag as imag,
    index_in_dim as index_in_dim,
    index_take as index_take,
    integer_pow as integer_pow,
    iota as iota,
    is_finite as is_finite,
    le as le,
    lgamma as lgamma,
    log as log,
    log1p as log1p,
    logistic as logistic,
    lt as lt,
    max as max,
    min as min,
    mul as mul,
    ne as ne,
    neg as neg,
    nextafter as nextafter,
    pad as pad,
    polygamma as polygamma,
    population_count as population_count,
    pow as pow,
    random_gamma_grad as random_gamma_grad,
    real as real,
    reciprocal as reciprocal,
    reduce as reduce,
    reduce_precision as reduce_precision,
    reduce_window as reduce_window,
    rem as rem,
    reshape as reshape,
    rev as rev,
    rng_bit_generator as rng_bit_generator,
    rng_uniform as rng_uniform,
    round as round,
    rsqrt as rsqrt,
    scatter as scatter,
    scatter_add as scatter_add,
    scatter_apply as scatter_apply,
    scatter_max as scatter_max,
    scatter_min as scatter_min,
    scatter_mul as scatter_mul,
    shift_left as shift_left,
    shift_right_arithmetic as shift_right_arithmetic,
    shift_right_logical as shift_right_logical,
    sign as sign,
    sin as sin,
    sinh as sinh,
    slice as slice,
    slice_in_dim as slice_in_dim,
    sort as sort,
    sort_key_val as sort_key_val,
    sqrt as sqrt,
    square as square,
    squeeze as squeeze,
    sub as sub,
    tan as tan,
    tanh as tanh,
    top_k as top_k,
    transpose as transpose,
    zeros_like_array as zeros_like_array,
    zeta as zeta,
)

# ----- Control Flow Operators -----
# isort: split
from jax.lax import (
    associative_scan as associative_scan,
    cond as cond,
    fori_loop as fori_loop,
    map as map,
    scan as scan,
    select as select,
    select_n as select_n,
    switch as switch,
    while_loop as while_loop,
)

# ----- Custom Gradient Operators -----
# isort: split
from jax.lax import (
    custom_linear_solve as custom_linear_solve,
    custom_root as custom_root,
    stop_gradient as stop_gradient,
)

# ----- Parallel Operators -----
# isort: split
from jax.lax import (
    all_gather as all_gather,
    all_to_all as all_to_all,
    axis_index as axis_index,
    pmax as pmax,
    pmean as pmean,
    pmin as pmin,
    ppermute as ppermute,
    pshuffle as pshuffle,
    psum as psum,
    psum_scatter as psum_scatter,
    pswapaxes as pswapaxes,
)

# ----- Sharding-related Operators -----
# isort: split
from jax.lax import with_sharding_constraint as with_sharding_constraint

# ----- Linear Algebra Operators -----
# isort: split
from jax.lax import linalg as linalg

# ----- Argument classes -----
# isort: split
from jax.lax import (
    ConvDimensionNumbers as ConvDimensionNumbers,
    ConvGeneralDilatedDimensionNumbers as ConvGeneralDilatedDimensionNumbers,
    DotAlgorithm as DotAlgorithm,
    GatherDimensionNumbers as GatherDimensionNumbers,
    GatherScatterMode as GatherScatterMode,
    Precision as Precision,
    PrecisionLike as PrecisionLike,
    RandomAlgorithm as RandomAlgorithm,
    RoundingMethod as RoundingMethod,
    ScatterDimensionNumbers as ScatterDimensionNumbers,
)
